"""
Sarmat.
Валидаторы.
Класс валидации объектов.
"""
from typing import Callable, TypeVar
from sarmat.core.context.models import BaseModel
from sarmat.core.constants.exception_constants import SarmatNotFilledAttribute

from .structure import ValidationIssue
from ..core.constants import ErrorClass

SarmatModel = TypeVar("SarmatModel", bound=BaseModel)


class Validator:
    """Класс для организации проверки Sarmat объектов."""

    def __init__(self, validate_call: Callable):
        self._call = validate_call
        self._issues: list[ValidationIssue] = []

    @property
    def issues(self) -> list[ValidationIssue]:
        """Ошибки валидации модели."""
        return self._issues

    def is_valid(self, model: SarmatModel, raise_error: bool = False) -> bool:
        """Проверка модели."""
        self._issues = self._call(model)
        is_valid = not self._issues

        if not is_valid and raise_error:
            raise SarmatNotFilledAttribute(
                err_class=ErrorClass.DATA,
                title="Ошибка валидации рейса",
                description=f"Замечаний к модели: {len(self._issues)}",
            )

        return is_valid



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Инструменты для построения расписания.
"""
from datetime import date, datetime, timedelta
from itertools import cycle
from uuid import uuid4

from sarmat.core.constants import DurationMonthCalcStrategy, SarmatWrongValueError
from sarmat.core.context.models import (
    CalendarPublication,
    IntervalModel,
    JourneyModel,
    JourneyPublication,
    RoutePublication,
)
from sarmat.core.context.tools.calculators.interval import get_datetime_sequence_from_interval_item
from sarmat.core.context.tools.calculators.price import PriceCalculator


def make_publication_from_journey(journey: JourneyModel, owner: int) -> JourneyPublication:
    """Создание объекта публикации рейса для станционного расписания.

    Args:
        journey: рейс
        owner: владелец публикации

    Returns: публикация
    """
    if not journey.departure_station:
        raise SarmatWrongValueError(description="Для рейса не указана станция отправления")

    if not journey.number:
        raise SarmatWrongValueError(description="Не указан номер рейса")

    uid = uuid4()
    route = [
        RoutePublication(
            journey=uid,
            sequence_number=0,
            point_id=journey.start_point.id,
            point_name=journey.start_point.name,
            station_id=journey.departure_station.id,
            station_name=journey.departure_station.name,
            departure_time=journey.departure_time,
        ),
    ]
    dt = datetime.combine(date.today(), journey.departure_time)

    for idx, item in enumerate(journey.structure):
        dt += timedelta(minutes=item.travel_time_min)
        arr_time = dt.time()

        if item.stop_time_min:
            dt += timedelta(minutes=item.stop_time_min)

        dep_time = dt.time()

        route.append(
            RoutePublication(
                journey=uid,
                sequence_number=idx+1,
                point_id=item.point.id,
                point_name=item.point.name,
                station_id=item.station.id if item.station else None,
                station_name=item.station.name if item.station else None,
                arrive_time=arr_time,
                departure_time=dep_time,
                stop_time=item.stop_time_min,
            )
        )

    return JourneyPublication(
        uid=uid,
        owner=owner,
        journey_type=journey.journey_type,
        name=journey.name,
        number=journey.number,
        literal=journey.literal,
        departure_time=journey.departure_time,
        is_chartered=journey.is_chartered,
        need_control=journey.need_control,
        route=route,
        calendar=[],
    )


def get_fare_matrix_for_journey(
    journey: JourneyModel,
    calc: PriceCalculator,
    sparce: bool = True
) -> list[list[float]]:
    """Вычисление матрицы стоимости проезда.

    Args:
        journey: рейс
        calc: правила вычисления стоимости
        sparce: признак для получения разреженных значений,
                когда каждая строка содержит полный набор значений,
                но для предыдущих пунктов эти значения равны 0

    Returns: последовательность значений для каждого пункта маршрута
    """
    route_len = len(journey.structure) + 1
    if sparce:
        fare = [[0.0 for _ in range(route_len)] for _ in range(route_len)]
    else:
        fare = []

    for i in range(route_len):
        if not sparce:
            fare.append([])

        for j in range(i + 1, route_len):
            val = calc.calculate_price(
                journey,
                departure_index=i,
                destination_index=j,
            )
            if sparce:
                fare[i][j] = round(val, 2)
            else:
                fare[i].append(round(val, 2))
    return fare


def fill_journey_price(
    publication: JourneyPublication,
    journey: JourneyModel,
    p_calculator: PriceCalculator,
    b_calculator: PriceCalculator,
) -> JourneyPublication:
    """Заполнение публикации сведениями о стоимости проезда

    Args:
        publication: публикуемое расписание для рейса
        journey: рейс
        p_calculator: правила расчёта стоимости проезда
        b_calculator: правила расчёта стоимости провоза багажа

    Returns: публикуемое расписание для рейса
    """
    pas_fare = get_fare_matrix_for_journey(journey, p_calculator, sparce=False)
    bag_fare = get_fare_matrix_for_journey(journey, b_calculator, sparce=False)

    for idx, r in enumerate(publication.route):
        r.ticket_amount = pas_fare[idx]
        r.baggage_amount = bag_fare[idx]

    return publication


def fill_journey_calendar(
    publication: JourneyPublication,
    journey: JourneyModel,
    interval: IntervalModel,
    operators: list[int],
    season_start: date,
    season_stop: date,
    calc_strategy: DurationMonthCalcStrategy,
    filter_by: int | None = None,
) -> JourneyPublication:
    """Заполнение графика выезда в рейс.

    Args:
        publication: публикация рейса
        journey: рейс
        interval: интервал выезда
        operators: перевозчики, обслуживающие рейс
        season_start: начало действия
        season_stop: окончание действия
        calc_strategy: стратегия расчёта времени для месяцев
        filter_by: формирование расписание по одному
                   из перевозчиков с учётом очерёдности обслуживания

    Returns: публикация рейса
    """
    operators_ordering = cycle(operators)
    if not interval.values:
        raise SarmatWrongValueError(description="Для интервала не указаны значения")

    datetime_sequence = get_datetime_sequence_from_interval_item(
        interval.values[0],
        datetime.combine(season_start, journey.departure_time),
        datetime.combine(season_stop, journey.departure_time),
        calc_strategy,
    )
    for i in datetime_sequence:
        operator = next(operators_ordering)
        if filter_by and operator != filter_by:
            continue

        publication.calendar.append(
            CalendarPublication(
                journey=publication.uid,
                carrier=operator,
                departure_date=i.date(),
            )
        )
    return publication



"""
Sarmat.
Валидаторы.
Проверка моделей перед публикацией рейса.
"""
from sarmat.core.context.models import (
    CalendarPublication,
    JourneyPublication,
    RoutePublication,
)

from .structure import ValidationIssue


def validate_calendar_publication(calendar: CalendarPublication) -> list[ValidationIssue]:
    """Проверка графика выезда."""
    issues = []

    if not calendar.journey:
        issues.append(ValidationIssue("journey", "Не указан идентификатор рейса"))

    if not calendar.carrier:
        issues.append(ValidationIssue("carrier", "Не указан перевозчик"))

    if not calendar.departure_date:
        issues.append(ValidationIssue("departure_date", "Не указана дата отправления"))

    return issues


def validate_route_item_publication(route_item: RoutePublication) -> list[ValidationIssue]:
    """Проверка элемента из состава маршрута."""
    issues = []

    if not route_item.journey:
        issues.append(ValidationIssue("journey", "Не указан идентификатор рейса"))

    if route_item.sequence_number is None or route_item.sequence_number < 0:
        issues.append(ValidationIssue("sequence_number", "Последовательность не указана или имеет неверное значение"))

    if not route_item.point_id:
        issues.append(ValidationIssue("point_id", "Не указан идентификатор пункта"))

    if not route_item.point_name:
        issues.append(ValidationIssue("point_name", "Не указано наименование пункта"))

    if route_item.station_id and not route_item.station_name:
        issues.append(ValidationIssue("station_name", "Не указано наименование станции"))

    if route_item.station_name and not route_item.station_id:
        issues.append(ValidationIssue("station_id", "Не указан идентификатор станции"))

    if route_item.sequence_number:
        if not route_item.arrive_time:
            issues.append(ValidationIssue("arrive_time", "Не указано время прибытия в пункт"))

    return issues


def validate_journey_publication(journey: JourneyPublication) -> list[ValidationIssue]:
    """Проверка рейса перед публикацией."""
    issues = []

    if not journey.uid:
        issues.append(ValidationIssue("uid", "Нет идентификатора рейса"))

    if not journey.owner:
        issues.append(ValidationIssue("owner", "Не указан автор публикации"))

    if not journey.journey_type:
        issues.append(ValidationIssue("journey_type", "Не указан тип рейса"))

    if not journey.name:
        issues.append(ValidationIssue("name", "Не указано наименование рейса"))

    if not journey.number:
        issues.append(ValidationIssue("number", "Не указан номер рейса"))

    if journey.is_chartered is None:
        issues.append(ValidationIssue("is_chartered", "Не указан признак заказного рейса"))

    if journey.need_control is None:
        issues.append(ValidationIssue("need_control", "Не указан признак паспортного контроля"))

    if not journey.departure_time:
        issues.append(ValidationIssue("departure_time", "Не указано время отправления"))

    if not journey.route:
        issues.append(ValidationIssue("route", "Нет состава маршрута"))

    for route_item in journey.route:
        issues.extend(validate_route_item_publication(route_item))
        if route_item.journey != journey.uid:
            issues.append(ValidationIssue("route_item", "Идентификатор точки маршрута не соответствует рейсу"))

    if not journey.calendar:
        issues.append(ValidationIssue("calendar", "Не указан график выезда"))

    for day in journey.calendar:
        issues.extend(validate_calendar_publication(day))
        if day.journey != journey.uid:
            issues.append(ValidationIssue("journey", "Идентификатор графика не соответствует рейсу"))

    if not journey.vehicle_model:
        issues.append(ValidationIssue("vehicle_model", "Не указана модель ТС"))

    if journey.seats and not journey.seats_map:
        issues.append(ValidationIssue("seats_map", "Не указана схема расположения мест"))

    if journey.seats_map and not journey.seats:
        issues.append(ValidationIssue("seats", "Не указано количество мест для сидения"))

    return issues



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
Объекты для работы со станционным расписанием.
"""
from dataclasses import dataclass
from datetime import date, time
from uuid import UUID

from .sarmat_models import BaseModel
from .traffic_management_models import JourneyType


@dataclass
class BaseRoutePublication(BaseModel):
    """Маршрут (базовый класс)."""

    journey: UUID           # идентификатор публикации рейса
    sequence_number: int    # номер в последовательности
    point_name: str         # наименование пункта
    point_id: int           # идентификатор пункта


@dataclass
class RoutePublication(BaseRoutePublication):
    """Маршрут."""

    station_name: str | None = None             # наименование станции
    station_id: int | None = None               # идентификатор станции
    arrive_time: time | None = None             # время прибытия в пункт
    stop_time: int | None = None                # время стоянки
    departure_time: time | None = None          # время отправления из пункта
    ticket_amount: list[float] | None = None    # стоимость проезда с текущего пункта до конца
    baggage_amount: list[float] | None = None   # стоимость провоза багажа с текущего пункта до конца


@dataclass
class CalendarPublication(BaseModel):
    """График выезда в рейс."""

    journey: UUID           # идентификатор рейса
    carrier: int            # перевозчик
    departure_date: date    # дата выезда


@dataclass
class BaseJourneyPublication(BaseModel):
    """Публикация рейса (основные атрибуты)."""

    uid: UUID                               # идентификатора записи
    owner: int                              # владелец записи
    journey_type: JourneyType               # тип рейса
    name: str                               # наименование рейса
    number: int                             # номер
    is_chartered: bool                      # заказной рейс
    need_control: bool                      # паспортный контроль
    departure_time: time                    # время отправления из начальной точки
    route: list[RoutePublication]           # маршрут
    calendar: list[CalendarPublication]     # график выезда в рейс


@dataclass
class JourneyPublication(BaseJourneyPublication):
    """Публикация рейса."""

    literal: str | None = None              # литера
    vehicle_model: str | None = None        # модель ТС
    baggage_capacity: int | None = None     # вместимость багажного отделения
    seats: int | None = None                # места для сидения
    stand_places: int | None = None         # места для проезда стоя
    seats_map: dict | None = None           # схема расположения мест



"""
Sarmat.
Валидаторы.
Структура для описания ошибок проверки.
"""
from dataclasses import dataclass


@dataclass
class ValidationIssue:
    """Описание проблемы."""

    field: str
    description: str



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
Вычисление интервала.
"""
from datetime import (
    datetime,
    timedelta,
)
from enum import Enum
from typing import Generator

from sarmat.core.constants import (
    DurationMonthCalcStrategy,
    IntervalType, SarmatWrongValueError,
)
from sarmat.core.context.models import IntervalItemModel

from .const import months_in_year
from .duration import get_month_increase


class LoopType(Enum):
    """Типы зацикленных значений."""

    DAY = "DAY"
    DAY_OF_WEEK = "DOW"


def get_timedelta_for_day(
    interval_value: int,
    calc_strategy: DurationMonthCalcStrategy,
) -> Generator[timedelta, None, None]:
    """Получение приращения по дате для помесячной генерации.

    Args:
        interval_value: значение для приращения
        calc_strategy: стратегия расчёта по месяцам

    Send args:
        start_datetime (datetime): момент отсчёта

    Yields: приращение по времени
    """
    res = timedelta(seconds=0)
    first_is_requested = False

    while True:
        start_datetime = yield res
        if start_datetime is None:
            raise SarmatWrongValueError(description="Не указано начало интервала")

        if not first_is_requested:
            first_is_requested = True
            continue

        val = get_month_increase(
            interval_value,
            start_datetime.date(),
            calc_strategy,
        )
        args = {IntervalType.DAY.as_cypher: val}
        res = timedelta(**args)


def get_timedelta_from_interval(
    interval_item: IntervalItemModel,
) -> Generator[timedelta, None, None]:
    """Получение простого приращения по дате.

    Args:
        interval_item: элемент интервала

    Send args:
        start_datetime (datetime): момент отсчёта

    Yields: приращение по времени
    """
    res = timedelta(seconds=0)
    args = {interval_item.interval_type.as_cypher: interval_item.values[0]}
    first_is_requested = False

    while True:
        _ = yield res
        if not first_is_requested:
            first_is_requested = True
            continue

        res = timedelta(**args)


def get_even_odd_timedelta(
    for_even: bool,
) -> Generator[timedelta, None, None]:
    """Получение приращения для чётных-нечётных чередований

    Args:
        for_even: признак расчёта для чётных дней

    Send args:
        start_datetime (datetime): момент отсчёта

    Yields: приращение по времени
    """
    res = timedelta(seconds=0)
    first_is_requested = False

    while True:
        start_datetime = yield res
        if start_datetime is None:
            raise SarmatWrongValueError(description="Не указано начало интервала")

        start_day = start_datetime.day

        # по чётным числам
        if for_even:
            if start_day % 2:
                val = 1
            else:
                val = 2
                if not first_is_requested:
                    first_is_requested = True
                    continue
        # по нечётным числам
        else:
            if start_day % 2:
                val = 2
                if not first_is_requested:
                    first_is_requested = True
                    continue
            else:
                val = 1

        first_is_requested = True
        args = {IntervalType.DAY.as_cypher: val}
        res = timedelta(**args)


def get_looped_sequence_timedelta(
    loop_type: LoopType,
    sequence: list[int],
    end_datetime: datetime,
) -> Generator[timedelta, None, None]:
    """Получение приращения для повторяющихся последовательностей

    Args:
        loop_type: тип последовательности
        sequence: набор значений
        end_datetime: дата окончания периода

    Send args:
        start_datetime (datetime): момент отсчёта

    Yields: приращение по времени
    """

    def get_calculated_day(dt_start: datetime) -> int:
        """Определение дня для вычисления.

        Args:
            dt_start: момент отсчёта времени

        Returns: высчитанный день недели или календарное число
        """
        if loop_type == LoopType.DAY_OF_WEEK:
            return dt_start.isoweekday()
        return dt_start.day

    res = timedelta(seconds=0)
    first_is_requested = False
    while True:
        delta = 0
        start_datetime = yield res
        if start_datetime is None:
            raise SarmatWrongValueError(description="Не указано начало интервала")

        if first_is_requested:
            delta += 1
            start_datetime += timedelta(days=1)

        days = (end_datetime - start_datetime).days

        for i in range(days + 1):
            calc_day = get_calculated_day(start_datetime + timedelta(days=i))
            if calc_day in sequence:
                break
            delta += 1
        else:
            break

        first_is_requested = True
        res = timedelta(days=delta)


def get_datetime_sequence_from_interval_item(
    interval_item: IntervalItemModel,
    start_datetime: datetime,
    stop_datetime: datetime,
    calc_strategy: DurationMonthCalcStrategy,
) -> Generator[datetime, None, None]:
    """Преобразование элемента из интервала в последовательность дат.

    Args:
        interval_item: элемент интервала
        start_datetime: начальный момент времени
        stop_datetime: конечный момент времени
        calc_strategy: стратегия расчёта времени для месяцев

    Yields: даты с приращением по интервалу
    """
    if interval_item.interval_type == IntervalType.MONTH:
        gen_timedelta = get_timedelta_for_day(
            interval_item.values[0],
            calc_strategy,
        )
    elif interval_item.interval_type == IntervalType.YEAR:
        gen_timedelta = get_timedelta_for_day(
            interval_item.values[0] * months_in_year,
            calc_strategy,
        )
    elif interval_item.interval_type == IntervalType.EVEN:
        gen_timedelta = get_even_odd_timedelta(for_even=True)
    elif interval_item.interval_type == IntervalType.ODD:
        gen_timedelta = get_even_odd_timedelta(for_even=False)
    elif interval_item.interval_type == IntervalType.DAYS_OF_MONTH:
        gen_timedelta = get_looped_sequence_timedelta(
            LoopType.DAY,
            interval_item.values,
            end_datetime=stop_datetime,
        )
    elif interval_item.interval_type == IntervalType.DAYS_OF_WEEK:
        gen_timedelta = get_looped_sequence_timedelta(
            LoopType.DAY_OF_WEEK,
            interval_item.values,
            end_datetime=stop_datetime,
        )
    else:
        gen_timedelta = get_timedelta_from_interval(interval_item)

    next(gen_timedelta)

    next_step = start_datetime
    while next_step <= stop_datetime:
        try:
            next_step += gen_timedelta.send(next_step)  # type: ignore [arg-type]
        except StopIteration:
            break

        if next_step <= stop_datetime:
            yield next_step
